# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io
import os
import platform
import re
import shutil
import subprocess
import tarfile
import tempfile

import pyomo.common.unittest as unittest
import pyomo.common.envvar as envvar

from pyomo.common import DeveloperError
from pyomo.common.fileutils import this_file, Executable
from pyomo.common.download import FileDownloader, distro_available
from pyomo.common.log import LoggingIntercept
from pyomo.common.tee import capture_output


class Test_FileDownloader(unittest.TestCase):
    def setUp(self):
        self.tmpdir = None
        self.basedir = os.path.abspath(os.path.curdir)

    def tearDown(self):
        if self.tmpdir:
            shutil.rmtree(self.tmpdir)

    def test_init(self):
        f = FileDownloader()
        self.assertFalse(f.insecure)
        self.assertIsNone(f.cacert)
        self.assertIsNone(f._fname)

        f = FileDownloader(True)
        self.assertTrue(f.insecure)
        self.assertIsNone(f.cacert)
        self.assertIsNone(f._fname)

        f = FileDownloader(True, this_file())
        self.assertTrue(f.insecure)
        self.assertEqual(f.cacert, this_file())
        self.assertIsNone(f._fname)

        with self.assertRaisesRegex(
            RuntimeError,
            "cacert='nonexistent_file_name' does not refer to a valid file.",
        ):
            FileDownloader(True, 'nonexistent_file_name')

    def test_parse(self):
        f = FileDownloader()
        f.parse_args([])
        self.assertFalse(f.insecure)
        self.assertIsNone(f.cacert)
        self.assertIsNone(f.target)

        f = FileDownloader()
        f.parse_args(['--insecure'])
        self.assertTrue(f.insecure)
        self.assertIsNone(f.cacert)
        self.assertIsNone(f.target)

        f = FileDownloader()
        f.parse_args(['--insecure', '--cacert', this_file()])
        self.assertTrue(f.insecure)
        self.assertEqual(f.cacert, this_file())
        self.assertIsNone(f.target)

        f = FileDownloader()
        f.parse_args(['--insecure', 'bar', '--cacert', this_file()])
        self.assertTrue(f.insecure)
        self.assertEqual(f.cacert, this_file())
        self.assertEqual(f.target, 'bar')

        f = FileDownloader()
        with capture_output() as io:
            with self.assertRaises(SystemExit):
                f.parse_args(['--cacert'])
        self.assertIn('argument --cacert: expected one argument', io.getvalue())

        f = FileDownloader()
        with capture_output() as io:
            with self.assertRaises(SystemExit):
                f.parse_args(['--cacert', '--insecure'])
        self.assertIn('argument --cacert: expected one argument', io.getvalue())

        f = FileDownloader()
        with self.assertRaisesRegex(
            RuntimeError,
            "--cacert='nonexistent_file_name' does not refer to a valid file",
        ):
            f.parse_args(['--cacert', 'nonexistent_file_name'])

        f = FileDownloader()
        with capture_output() as io:
            with self.assertRaises(SystemExit):
                f.parse_args(['--foo'])
        self.assertIn('error: unrecognized arguments: --foo', io.getvalue())

    def test_set_destination_filename(self):
        self.tmpdir = os.path.abspath(tempfile.mkdtemp())

        f = FileDownloader()
        self.assertIsNone(f._fname)
        f.set_destination_filename('foo')
        self.assertEqual(f._fname, os.path.join(envvar.PYOMO_CONFIG_DIR, 'foo'))
        # By this point, the CONFIG_DIR is guaranteed to have been created
        self.assertTrue(os.path.isdir(envvar.PYOMO_CONFIG_DIR))

        f.target = self.tmpdir
        f.set_destination_filename('foo')
        target = os.path.join(self.tmpdir, 'foo')
        self.assertEqual(f._fname, target)
        self.assertFalse(os.path.exists(target))

        f.target = self.tmpdir
        f.set_destination_filename(os.path.join('foo', 'bar'))
        target = os.path.join(self.tmpdir, 'foo', 'bar')
        self.assertEqual(f._fname, target)
        self.assertFalse(os.path.exists(target))
        target_dir = os.path.join(self.tmpdir, 'foo')
        self.assertTrue(os.path.isdir(target_dir))

    def test_get_sysinfo(self):
        f = FileDownloader()
        ans = f.get_sysinfo()
        self.assertIs(type(ans), tuple)
        self.assertEqual(len(ans), 2)
        self.assertTrue(len(ans[0]) > 0)
        self.assertTrue(platform.system().lower().startswith(ans[0]))
        self.assertFalse(any(c in ans[0] for c in '.-_'))
        self.assertIn(ans[1], (32, 64))

    def test_get_os_version(self):
        f = FileDownloader()
        _os, _ver = f.get_os_version(normalize=False)
        _norm = f.get_os_version(normalize=True)
        # print(_os,_ver,_norm)
        _sys = f.get_sysinfo()[0]
        if _sys == 'linux':
            dist, dist_ver = re.match('^([^0-9]+)(.*)', _norm).groups()
            self.assertNotIn('.', dist_ver)
            self.assertGreater(int(dist_ver), 0)
            if dist == 'ubuntu':
                self.assertEqual(dist_ver, ''.join(_ver.split('.')[:2]))
            else:
                self.assertEqual(dist_ver, _ver.split('.')[0])

            if distro_available:
                d, v = f._get_distver_from_distro()
                # print(d,v)
                self.assertEqual(_os, d)
                self.assertEqual(_ver, v)
                self.assertTrue(v.replace('.', '').startswith(dist_ver))

            if os.path.exists('/etc/redhat-release'):
                d, v = f._get_distver_from_redhat_release()
                # print(d,v)
                self.assertEqual(_os, d)
                self.assertEqual(_ver, v)
                self.assertTrue(v.replace('.', '').startswith(dist_ver))

            if (
                Executable('lsb_release').available()
                and subprocess.run(
                    ['lsb_release'],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                ).returncode
                == 0
            ):
                d, v = f._get_distver_from_lsb_release()
                # print(d,v)
                self.assertEqual(_os, d)
                self.assertEqual(_ver, v)
                self.assertTrue(v.replace('.', '').startswith(dist_ver))

            if os.path.exists('/etc/os-release'):
                d, v = f._get_distver_from_os_release()
                # print(d,v)
                self.assertEqual(_os, d)
                # Note that (at least on centos), os_release is an
                # imprecise version string
                self.assertTrue(_ver.startswith(v))
                self.assertTrue(v.replace('.', '').startswith(dist_ver))

        elif _sys == 'darwin':
            dist, dist_ver = re.match('^([^0-9]+)(.*)', _norm).groups()
            self.assertEqual(_os, 'macos')
            self.assertEqual(dist, 'macos')
            self.assertNotIn('.', dist_ver)
            self.assertGreater(int(dist_ver), 0)
            self.assertEqual(_norm, _os + ''.join(_ver.split('.')[:2]))
        elif _sys == 'windows':
            self.assertEqual(_os, 'win')
            self.assertEqual(_norm, _os + ''.join(_ver.split('.')[:2]))
        else:
            self.assertEqual(_os, '')

        self.assertEqual((_os, _ver), FileDownloader._os_version)
        # Exercise the fetch from CACHE
        try:
            FileDownloader._os_version, tmp = ("test", '2'), FileDownloader._os_version
            self.assertEqual(f.get_os_version(False), ("test", "2"))
            self.assertEqual(f.get_os_version(), "test2")
        finally:
            FileDownloader._os_version = tmp

    def test_get_platform_url(self):
        f = FileDownloader()
        urlmap = {'bogus_sys': 'bogus'}
        with self.assertRaisesRegex(
            RuntimeError, "cannot infer the correct url for platform '.*'"
        ):
            f.get_platform_url(urlmap)

        urlmap[f.get_sysinfo()[0]] = 'correct'
        self.assertEqual(f.get_platform_url(urlmap), 'correct')

    def test_get_files_requires_set_destination(self):
        f = FileDownloader()
        with self.assertRaisesRegex(
            DeveloperError, 'target file name has not been initialized'
        ):
            f.get_binary_file('bogus')

        with self.assertRaisesRegex(
            DeveloperError, 'target file name has not been initialized'
        ):
            f.get_binary_file_from_zip_archive('bogus', 'bogus')

        with self.assertRaisesRegex(
            DeveloperError, 'target file name has not been initialized'
        ):
            f.get_gzipped_binary_file('bogus')

    def test_get_text_binary_file(self):
        tmpdir = tempfile.mkdtemp()
        try:
            f = FileDownloader()

            # Mock retrieve_url so network connections are not necessary
            f.retrieve_url = lambda url: bytes("\n", encoding='utf-8')

            # Binary files will preserve line endings
            target = os.path.join(tmpdir, 'bin.txt')
            f.set_destination_filename(target)
            f.get_binary_file(None)
            self.assertEqual(os.path.getsize(target), 1)

            # Text files will convert line endings to the local platform
            target = os.path.join(tmpdir, 'txt.txt')
            f.set_destination_filename(target)
            f.get_text_file(None)
            self.assertEqual(os.path.getsize(target), len(os.linesep))
        finally:
            shutil.rmtree(tmpdir)

    def test_get_tar_archive(self):
        tmpdir = tempfile.mkdtemp()
        try:
            f = FileDownloader()

            # Mock retrieve_url so network connections are not necessary
            buf = io.BytesIO()
            with tarfile.open(mode="w:gz", fileobj=buf) as TAR:
                info = tarfile.TarInfo('b/lnk')
                info.size = 0
                info.type = tarfile.SYMTYPE
                info.linkname = envvar.PYOMO_CONFIG_DIR
                TAR.addfile(info)
                for fname in ('a', 'b/c', 'b/d', '/root', 'b/lnk/test'):
                    info = tarfile.TarInfo(fname)
                    info.size = 0
                    info.type = tarfile.REGTYPE
                    info.mode = 0o644
                    info.mtime = info.uid = info.gid = 0
                    info.uname = info.gname = 'root'
                    TAR.addfile(info)
            f.retrieve_url = lambda url: buf.getvalue()

            with self.assertRaisesRegex(
                DeveloperError,
                r"(?s)target file name has not been initialized "
                r"with set_destination_filename".replace(' ', r'\s+'),
            ):
                f.get_tar_archive(None, 1)

            _tmp = os.path.join(tmpdir, 'a_file')
            with open(_tmp, 'w'):
                pass
            f.set_destination_filename(_tmp)
            with self.assertRaisesRegex(
                RuntimeError,
                r"Target directory \(.*a_file\) exists, but is not a directory",
            ):
                f.get_tar_archive(None, 1)

            f.set_destination_filename(tmpdir)
            with LoggingIntercept() as LOG:
                f.get_tar_archive(None, 1)

            self.assertEqual(
                LOG.getvalue().strip(),
                """
Skipping file (a) in tar archive due to dirOffset.
malformed or potentially insecure filename (/root).  Skipping file.
potentially insecure filename (lnk/test) resolves outside target directory.  Skipping file.
""".strip(),
            )
            for f in ('c', 'd'):
                fname = os.path.join(tmpdir, f)
                self.assertTrue(os.path.exists(fname))
                self.assertTrue(os.path.isfile(fname))
            for f in ('lnk',):
                fname = os.path.join(tmpdir, f)
                self.assertTrue(os.path.exists(fname))
                self.assertTrue(os.path.islink(fname))
        finally:
            shutil.rmtree(tmpdir)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import functools
import inspect
import sys

from pyomo.common import DeveloperError

#
# The following decorator is general and should probably be promoted to
# component.py so that we can efficiently handle construction errors on
# scalar components.
#
# TODO: quantify the memory overhead here.  We create (and preserve) a
# locals() dict for *each* method that we wrap.  If that becomes
# significant, we might consider using a single global private
# environment (which would require some thought when managing any
# potential name collisions)
#
_disabled_error = (
    "Cannot %s %s '%s' before it has been constructed (initialized): "
    "'%s' is an attribute on an Abstract component and cannot be "
    "accessed until the component has been fully constructed (converted "
    "to a Concrete component) using AbstractModel.create_instance() or "
    "%s.construct()."
)


def _disable_method(fcn, msg=None, exception=RuntimeError):
    _name = fcn.__name__
    if msg is None:
        msg = "access '%s' on" % (_name,)

    # functools.wraps doesn't preserve the function signature until
    # Python 3.4, and even then, does not preserve it accurately (e.g.,
    # calling with the incorrect number of arguments does not generate
    # an error).  For backwards compatibility with Python 2.x, we will
    # create a temporary (local) function using exec that matches the
    # function signature passed in and raises an exception
    sig = inspect.signature(fcn)

    params = list(sig.parameters.values())
    for i, param in enumerate(params):
        if param.default is inspect.Parameter.empty:
            continue
        # The function that we are going to wrap needs to have the same
        # notional signature as the original function (positional args,
        # optional args, kwargs), but things like the default values do
        # not have to be the same (as functools.wraps will transfer over
        # defaults).  As some default values (types, local object
        # instances) do not transfer well when casting the signature to
        # a string, we will convert all optional positional arguments to
        # have a default value of None
        params[i] = param.replace(default=None)
    sig = sig.replace(parameters=params)
    args = str(sig)
    assert args == '(self)' or args.startswith('(self,')

    # lambda comes through with a function name "<lambda>".  We will
    # use exec here to create a function (in a private namespace)
    # that will have the correct name and signature.
    _env = {'_msg': msg, '_name': _name}
    _funcdef = """def %s%s:
        raise %s("%s" %% (_msg, type(self).__name__,
            self.name, _name, self.name))
""" % (_name, args, exception.__name__, _disabled_error)
    exec(_funcdef, _env)
    return functools.wraps(fcn)(_env[_name])


def _disable_property(fcn, msg=None, exception=RuntimeError):
    _name = fcn.fget.__name__
    if msg is None:
        _gmsg = "access property '%s' on" % (_name,)
    else:
        _gmsg = msg
    getter = _disable_method(fcn.fget, _gmsg, exception)

    if fcn.fset is None:
        setter = None
    else:
        if msg is None:
            _smsg = "set property '%s' on" % (_name,)
        else:
            _smsg = msg
        setter = _disable_method(fcn.fset, _smsg, exception)

    return property(fget=getter, fset=setter, doc=fcn.__doc__)


def disable_methods(methods):
    """Class decorator to disable methods before construct is called.

    This decorator should be called to create "Abstract" scalar classes
    that override key methods to raise exceptions.  When the construct()
    method is called, the class instance changes type back to the
    original scalar component and the full class functionality is
    restored.  This prevents most class methods from having to begin with
    "`if not self.parent_component()._constructed: raise RuntimeError`"
    """

    def class_decorator(cls):
        assert len(cls.__bases__) == 1
        base = cls.__bases__[0]

        def construct(self, data=None):
            if hasattr(self, '_name') and self._name == self.__class__.__name__:
                self._name = base.__name__
            self.__class__ = base
            return base.construct(self, data)

        construct.__doc__ = base.construct.__doc__
        cls.construct = construct

        for method in methods:
            msg = None
            exc = RuntimeError
            if type(method) is tuple:
                if len(method) == 2:
                    method, msg = method
                else:
                    method, msg, exc = method
            if not hasattr(base, method):
                raise DeveloperError(
                    "Cannot disable method %s on %s: not present on base class"
                    % (method, cls)
                )
            base_method = getattr(base, method)
            if type(base_method) is property:
                setattr(cls, method, _disable_property(base_method, msg, exc))
            else:
                setattr(cls, method, _disable_method(base_method, msg, exc))
        return cls

    return class_decorator



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.cp.interval_var import (
    IntervalVar,
    IntervalVarStartTime,
    IntervalVarEndTime,
    IntervalVarLength,
    IntervalVarPresence,
)
from pyomo.contrib.cp.repn.docplex_writer import DocplexWriter, CPOptimizerSolver
from pyomo.contrib.cp.sequence_var import SequenceVar
from pyomo.contrib.cp.scheduling_expr.sequence_expressions import (
    no_overlap,
    first_in_sequence,
    last_in_sequence,
    before_in_sequence,
    predecessor_to,
)
from pyomo.contrib.cp.scheduling_expr.scheduling_logic import (
    alternative,
    spans,
    synchronize,
)
from pyomo.contrib.cp.scheduling_expr.step_function_expressions import (
    AlwaysIn,
    Step,
    Pulse,
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.contrib.viewer.qt as myqt
from pyomo.contrib.viewer.pyomo_viewer import qtconsole_app, qtconsole_available


class QtApp(qtconsole_app.JupyterQtConsoleApp if qtconsole_available else object):
    _kernel_cmd_show_ui = """try:
    ui.show()
except NameError:
    try:
        model
    except NameError:
        model=None
    ui = get_mainwindow(model=model, ask_close=False)
ui.setWindowTitle('Pyomo Model Viewer -- {}')"""

    _kernel_cmd_hide_ui = """try:
    ui.hide()
except NameError:
    pass"""

    _kernel_cmd_import_qt_magic = r"%gui qt"

    _kernel_cmd_import_ui = "from pyomo.contrib.viewer.ui import get_mainwindow"

    _kernel_cmd_import_pyomo_env = "import pyomo.environ as pyo"

    def active_widget_name(self):
        current_widget = self.window.tab_widget.currentWidget()
        current_widget_index = self.window.tab_widget.indexOf(current_widget)
        return self.window.tab_widget.tabText(current_widget_index).replace("'", '"')

    def show_ui(self):
        kc = self.window.active_frontend.kernel_client
        kc.execute(
            self._kernel_cmd_show_ui.format(self.active_widget_name()), silent=True
        )

    def hide_ui(self):
        kc = self.window.active_frontend.kernel_client
        kc.execute(self._kernel_cmd_hide_ui, silent=True)

    def run_script(self, checked=False, filename=None):
        """Run a python script in the current kernel."""
        if filename is None:
            # Show a dialog box for user to select working directory
            filename = myqt.QtWidgets.QFileDialog.getOpenFileName(
                self.window,
                "Run Script",
                os.getcwd(),
                "py (*.py);;text (*.txt);;all (*)",
            )
            if filename[0]:  # returns a tuple of file and filter or ("","")
                filename = filename[0]
            else:
                filename = None
        # Run script if one was selected
        if filename is not None:
            kc = self.window.active_frontend.kernel_client
            kc.execute("%run {}".format(filename))

    def kernel_pyomo_init(self, kc):
        kc.execute(self._kernel_cmd_import_qt_magic, silent=True)
        kc.execute(self._kernel_cmd_import_ui, silent=True)
        kc.execute(self._kernel_cmd_import_pyomo_env, silent=False)

    def init_qt_elements(self):
        super().init_qt_elements()
        self.kernel_pyomo_init(self.widget.kernel_client)
        self.run_script_act = myqt.QAction("&Run Script...", self.window)
        self.show_ui_act = myqt.QAction("&Show Pyomo Model Viewer", self.window)
        self.hide_ui_act = myqt.QAction("&Hide Pyomo Model Viewer", self.window)
        self.window.file_menu.addSeparator()
        self.window.file_menu.addAction(self.run_script_act)
        self.window.view_menu.addSeparator()
        self.window.view_menu.addAction(self.show_ui_act)
        self.window.view_menu.addAction(self.hide_ui_act)
        self.window.view_menu.addSeparator()
        self.run_script_act.triggered.connect(self.run_script)
        self.show_ui_act.triggered.connect(self.show_ui)
        self.hide_ui_act.triggered.connect(self.hide_ui)

    def new_frontend_master(self):
        widget = super().new_frontend_master()
        self.kernel_pyomo_init(widget.kernel_client)
        return widget



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.common.collections import ComponentSet
from pyomo.core.base import Transformation, TransformationFactory
from pyomo.core import Var, Expression, Objective
from pyomo.dae import ContinuousSet, DerivativeVar, Integral

from pyomo.dae.misc import generate_finite_elements
from pyomo.dae.misc import expand_components
from pyomo.dae.misc import create_partial_expression
from pyomo.dae.misc import add_discretization_equations
from pyomo.dae.misc import block_fully_discretized
from pyomo.dae.diffvar import DAE_Error

from pyomo.common.config import ConfigBlock, ConfigValue, PositiveInt, In

logger = logging.getLogger('pyomo.dae')


def _central_transform(v, s):
    """
    Applies the Central Difference formula of order O(h^2) for first
    derivatives
    """

    def _ctr_fun(i):
        tmp = list(s)
        idx = s.ord(i) - 1
        if idx == 0:  # Needed since '-1' is considered a valid index in Python
            raise IndexError("list index out of range")
        return 1 / (tmp[idx + 1] - tmp[idx - 1]) * (v(tmp[idx + 1]) - v(tmp[idx - 1]))

    return _ctr_fun


def _central_transform_order2(v, s):
    """
    Applies the Central Difference formula of order O(h^2) for second
    derivatives
    """

    def _ctr_fun2(i):
        tmp = list(s)
        idx = s.ord(i) - 1
        if idx == 0:  # Needed since '-1' is considered a valid index in Python
            raise IndexError("list index out of range")
        return (
            1
            / ((tmp[idx + 1] - tmp[idx]) * (tmp[idx] - tmp[idx - 1]))
            * (v(tmp[idx + 1]) - 2 * v(tmp[idx]) + v(tmp[idx - 1]))
        )

    return _ctr_fun2


def _forward_transform(v, s):
    """
    Applies the Forward Difference formula of order O(h) for first derivatives
    """

    def _fwd_fun(i):
        tmp = list(s)
        idx = s.ord(i) - 1
        return 1 / (tmp[idx + 1] - tmp[idx]) * (v(tmp[idx + 1]) - v(tmp[idx]))

    return _fwd_fun


def _forward_transform_order2(v, s):
    """
    Applies the Forward Difference formula of order O(h) for second derivatives
    """

    def _fwd_fun(i):
        tmp = list(s)
        idx = s.ord(i) - 1
        return (
            1
            / ((tmp[idx + 2] - tmp[idx + 1]) * (tmp[idx + 1] - tmp[idx]))
            * (v(tmp[idx + 2]) - 2 * v(tmp[idx + 1]) + v(tmp[idx]))
        )

    return _fwd_fun


def _backward_transform(v, s):
    """
    Applies the Backward Difference formula of order O(h) for first derivatives
    """

    def _bwd_fun(i):
        tmp = list(s)
        idx = s.ord(i) - 1
        if idx == 0:  # Needed since '-1' is considered a valid index in Python
            raise IndexError("list index out of range")
        return 1 / (tmp[idx] - tmp[idx - 1]) * (v(tmp[idx]) - v(tmp[idx - 1]))

    return _bwd_fun


def _backward_transform_order2(v, s):
    """
    Applies the Backward Difference formula of order O(h) for second
    derivatives
    """

    def _bwd_fun(i):
        tmp = list(s)
        idx = s.ord(i) - 1

        # This check is needed since '-1' is considered a valid index in Python
        if idx == 0 or idx == 1:
            raise IndexError("list index out of range")
        return (
            1
            / ((tmp[idx - 1] - tmp[idx - 2]) * (tmp[idx] - tmp[idx - 1]))
            * (v(tmp[idx]) - 2 * v(tmp[idx - 1]) + v(tmp[idx - 2]))
        )

    return _bwd_fun


@TransformationFactory.register(
    'dae.finite_difference',
    doc="Discretizes a DAE model using "
    "a finite difference method transforming the model into an NLP.",
)
class Finite_Difference_Transformation(Transformation):
    """
    Transformation that applies finite difference methods to
    DAE, ODE, or PDE models.
    """

    CONFIG = ConfigBlock("dae.finite_difference")
    CONFIG.declare(
        'nfe',
        ConfigValue(
            default=10,
            domain=PositiveInt,
            description="The desired number of finite element points to be "
            "included in the discretization",
        ),
    )
    CONFIG.declare(
        'wrt',
        ConfigValue(
            default=None,
            description="The ContinuousSet to be discretized",
            doc="Indicates which ContinuousSet the transformation should be "
            "applied to. If this keyword argument is not specified then the "
            "same scheme will be applied to all ContinuousSets.",
        ),
    )
    CONFIG.declare(
        'scheme',
        ConfigValue(
            default='BACKWARD',
            domain=In(['BACKWARD', 'CENTRAL', 'FORWARD']),
            description="Indicates which finite difference scheme to apply",
            doc="Options are BACKWARD, CENTRAL, or FORWARD. The default scheme is "
            "the backward difference method",
        ),
    )

    def __init__(self):
        super(Finite_Difference_Transformation, self).__init__()
        self._nfe = {}
        self.all_schemes = {
            'BACKWARD': (_backward_transform, _backward_transform_order2),
            'CENTRAL': (_central_transform, _central_transform_order2),
            'FORWARD': (_forward_transform, _forward_transform_order2),
        }

    def _apply_to(self, instance, **kwds):
        """
        Applies the transformation to a modeling instance

        Keyword Arguments:
        nfe           The desired number of finite element points to be
                      included in the discretization.
        wrt           Indicates which ContinuousSet the transformation
                      should be applied to. If this keyword argument is not
                      specified then the same scheme will be applied to all
                      ContinuousSets.
        scheme        Indicates which finite difference method to apply.
                      Options are BACKWARD, CENTRAL, or FORWARD. The default
                      scheme is the backward difference method
        """

        config = self.CONFIG(kwds)

        tmpnfe = config.nfe
        tmpds = config.wrt

        if tmpds is not None:
            if tmpds.ctype is not ContinuousSet:
                raise TypeError(
                    "The component specified using the 'wrt' "
                    "keyword must be a continuous set"
                )
            elif 'scheme' in tmpds.get_discretization_info():
                raise ValueError(
                    "The discretization scheme '%s' has already "
                    "been applied to the ContinuousSet '%s'"
                    % (tmpds.get_discretization_info()['scheme'], tmpds.name)
                )

        if None in self._nfe:
            raise ValueError(
                "A general discretization scheme has already been applied to "
                "to every continuous set in the model. If you would like to "
                "apply a different discretization scheme to each continuous "
                "set, you must declare a new transformation object"
            )

        if len(self._nfe) == 0 and tmpds is None:
            # Same discretization on all ContinuousSets
            self._nfe[None] = tmpnfe
            currentds = None
        else:
            self._nfe[tmpds.name] = tmpnfe
            currentds = tmpds.name

        self._scheme_name = config.scheme
        self._scheme = self.all_schemes.get(self._scheme_name, None)

        self._transformBlock(instance, currentds)

    def _transformBlock(self, block, currentds):
        self._fe = {}
        for ds in block.component_objects(ContinuousSet):
            if currentds is None or currentds == ds.name or currentds is ds:
                if 'scheme' in ds.get_discretization_info():
                    raise DAE_Error(
                        "Attempting to discretize ContinuousSet "
                        "'%s' after it has already been discretized. " % ds.name
                    )
                generate_finite_elements(ds, self._nfe[currentds])
                if not ds.get_changed():
                    if len(ds) - 1 > self._nfe[currentds]:
                        logger.warning(
                            "More finite elements were found in "
                            "ContinuousSet '%s' than the number of "
                            "finite elements specified in apply. The "
                            "larger number of finite elements will be "
                            "used." % ds.name
                        )

                self._nfe[ds.name] = len(ds) - 1
                self._fe[ds.name] = list(ds)
                # Adding discretization information to the ContinuousSet
                # object itself so that it can be accessed outside of the
                # discretization object
                disc_info = ds.get_discretization_info()
                disc_info['nfe'] = self._nfe[ds.name]
                disc_info['scheme'] = self._scheme_name + ' Difference'

        # Maybe check to see if any of the ContinuousSets have been changed,
        # if they haven't then the model components need not be updated
        # or even iterated through
        expand_components(block)

        for d in block.component_objects(DerivativeVar, descend_into=True):
            dsets = d.get_continuousset_list()
            for i in ComponentSet(dsets):
                if currentds is None or i.name == currentds:
                    oldexpr = d.get_derivative_expression()
                    loc = d.get_state_var()._contset[i]
                    count = dsets.count(i)
                    if count >= 3:
                        raise DAE_Error(
                            "Error discretizing '%s' with respect to '%s'. "
                            "Current implementation only allows for taking the"
                            " first or second derivative with respect to "
                            "a particular ContinuousSet" % (d.name, i.name)
                        )
                    scheme = self._scheme[count - 1]
                    newexpr = create_partial_expression(scheme, oldexpr, i, loc)
                    d.set_derivative_expression(newexpr)

            # Reclassify DerivativeVar if all indexing ContinuousSets have
            # been discretized. Add discretization equations to the
            # DerivativeVar's parent block.
            if d.is_fully_discretized():
                add_discretization_equations(d.parent_block(), d)
                d.parent_block().reclassify_component_type(d, Var)

                # Keep track of any reclassified DerivativeVar components so
                # that the Simulator can easily identify them if the model
                # is simulated after discretization
                # TODO: Update the discretization transformations to use
                # a Block to add things to the model and store discretization
                # information. Using a list for now because the simulator
                # does not yet support models containing active Blocks
                reclassified_list = getattr(
                    block, '_pyomo_dae_reclassified_derivativevars', None
                )
                if reclassified_list is None:
                    block._pyomo_dae_reclassified_derivativevars = list()
                    reclassified_list = block._pyomo_dae_reclassified_derivativevars

                reclassified_list.append(d)

        # Reclassify Integrals if all ContinuousSets have been discretized
        if block_fully_discretized(block):
            if block.contains_component(Integral):
                for i in block.component_objects(Integral, descend_into=True):
                    i.parent_block().reclassify_component_type(i, Expression)
                    # TODO: The following reproduces the old behavior of
                    # "reconstruct()".  We should come up with an
                    # implementation that does not rely on manipulating
                    # private attributes
                    i.clear()
                    i._constructed = False
                    i.construct()

                # If a model contains integrals they are most likely to
                # appear in the objective function which will need to be
                # reconstructed after the model is discretized.
                for k in block.component_objects(Objective, descend_into=True):
                    # TODO: check this, reconstruct might not work
                    # TODO: The following reproduces the old behavior of
                    # "reconstruct()".  We should come up with an
                    # implementation that does not rely on manipulating
                    # private attributes
                    k.clear()
                    k._constructed = False
                    k.construct()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import os
import subprocess
import datetime
import time
import io
import re
import sys
import time
import threading
from typing import Optional, Tuple, Union, Mapping, List, Dict, Any, Sequence

from pyomo.common import Executable
from pyomo.common.config import (
    ConfigDict,
    ConfigList,
    ConfigValue,
    document_configdict,
    document_class_CONFIG,
    ADVANCED_OPTION,
)
from pyomo.common.errors import (
    ApplicationError,
    InfeasibleConstraintException,
    MouseTrap,
)
from pyomo.common.fileutils import to_legal_filename
from pyomo.common.tempfiles import TempfileManager
from pyomo.common.timing import HierarchicalTimer, default_timer
from pyomo.core.base.var import VarData
from pyomo.core.staleflag import StaleFlagManager
from pyomo.repn.plugins.nl_writer import NLWriter, NLWriterInfo
from pyomo.contrib.solver.common.base import SolverBase, Availability
from pyomo.contrib.solver.common.config import SolverConfig
from pyomo.contrib.solver.common.factory import LegacySolverWrapper
from pyomo.contrib.solver.common.results import (
    Results,
    TerminationCondition,
    SolutionStatus,
)
from pyomo.contrib.solver.solvers.asl_sol_reader import (
    asl_solve_code_to_solution_status,
    parse_asl_sol_file,
    ASLSolFileData,
    ASLSolFileSolutionLoader,
)
from pyomo.contrib.solver.common.util import NoOptimalSolutionError, NoSolutionError
from pyomo.common.tee import TeeStream
from pyomo.core.expr.visitor import replace_expressions
from pyomo.core.expr.numvalue import value
from pyomo.core.base.suffix import Suffix
from pyomo.common.collections import ComponentMap
from pyomo.solvers.amplfunc_merge import amplfunc_merge

logger = logging.getLogger(__name__)

# Acceptable chars for the end of the alpha_pr column
# in ipopt's output, per https://coin-or.github.io/Ipopt/OUTPUT.html
_ALPHA_PR_CHARS = set("fFhHkKnNRwSstTr")


def _option_to_cmd(opt: str, val: str | int | float):
    """Convert a option / value pair into a valid command line argument."""
    if isinstance(val, str):
        if '"' not in val:
            return f'{opt}="{val}"'
        elif "'" not in val:
            return f"{opt}='{val}'"
        else:
            raise ValueError(
                f"solver_option '{opt}' contained value {val!r} with "
                "both single and double quotes.  Ipopt cannot parse "
                "command line options with escaped quote characters."
            )
    else:
        return f'{opt}={val}'


@document_configdict()
class IpoptConfig(SolverConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.executable: Executable = self.declare(
            'executable',
            ConfigValue(
                domain=Executable,
                default='ipopt',
                description="Preferred executable for ipopt. Defaults to searching "
                "the ``PATH`` for the first available ``ipopt``.",
            ),
        )
        self.writer_config: ConfigDict = self.declare(
            'writer_config', NLWriter.CONFIG()
        )


class IpoptSolutionLoader(ASLSolFileSolutionLoader):
    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        if self._nl_info.eliminated_vars:
            raise MouseTrap(
                'Complete reduced costs are not available when variables have '
                'been presolved from the model.  Turn presolve off '
                '(solver.config.writer_config.linear_presolve=False) to get '
                'reduced costs.'
            )

        zl_map = self._sol_data.var_suffixes.get('ipopt_zL_out', {})
        zu_map = self._sol_data.var_suffixes.get('ipopt_zU_out', {})
        # TBD: is it an error if Ipopt fails to return RC info?
        # if not (zl_map or zu_map):
        #     raise?
        if self._nl_info.scaling:
            # Unscale the zl and zu maps:
            inv_obj_scale = 1.0
            if self._nl_info.scaling.objectives:
                inv_obj_scale /= self._nl_info.scaling.objectives[self._sol_data.objno]
            var_scale = self._nl_info.scaling.variables
            zl_map = {k: v * var_scale[k] * inv_obj_scale for k, v in zl_map.items()}
            zu_map = {k: v * var_scale[k] * inv_obj_scale for k, v in zu_map.items()}

        rc = ComponentMap()
        for ndx, v in enumerate(self._nl_info.variables):
            _rc = 0.0
            if ndx in zl_map:
                # Note *any* value in zl has an absolute value at least
                # as big as 0.  No need to test and just overwrite _rc:
                _rc = zl_map[ndx]
            if ndx in zu_map:
                zu = zu_map[ndx]
                if abs(zu) > abs(_rc):
                    _rc = zu
            rc[v] = _rc

        if vars_to_load is not None:
            # Note vars_to_load could contain variables that were
            # eliminated (so use get()):
            rc = ComponentMap((v, rc.get(v, 0)) for v in vars_to_load)
        return rc


#: The set of all ipopt options that can be passed to Ipopt on the command line
ipopt_command_line_options = {
    'acceptable_compl_inf_tol',
    'acceptable_constr_viol_tol',
    'acceptable_dual_inf_tol',
    'acceptable_tol',
    'alpha_for_y',
    'bound_frac',
    'bound_mult_init_val',
    'bound_push',
    'bound_relax_factor',
    'compl_inf_tol',
    'constr_mult_init_max',
    'constr_viol_tol',
    'diverging_iterates_tol',
    'dual_inf_tol',
    'expect_infeasible_problem',
    'file_print_level',
    'halt_on_ampl_error',
    'hessian_approximation',
    'honor_original_bounds',
    'linear_scaling_on_demand',
    'linear_solver',
    'linear_system_scaling',
    'ma27_pivtol',
    'ma27_pivtolmax',
    'ma57_pivot_order',
    'ma57_pivtol',
    'ma57_pivtolmax',
    'max_cpu_time',
    'max_iter',
    'max_refinement_steps',
    'max_soc',
    'maxit',
    'min_refinement_steps',
    'mu_init',
    'mu_max',
    'mu_oracle',
    'mu_strategy',
    'nlp_scaling_max_gradient',
    'nlp_scaling_method',
    'obj_scaling_factor',
    'option_file_name',
    'outlev',
    'output_file',
    'pardiso_matching_strategy',
    'print_level',
    'print_options_documentation',
    'print_user_options',
    'required_infeasibility_reduction',
    'slack_bound_frac',
    'slack_bound_push',
    'tol',
    'wantsol',
    'warm_start_bound_push',
    'warm_start_init_point',
    'warm_start_mult_bound_push',
    'watchdog_shortened_iter_trigger',
}

#: The set of options we forbid the user from setting (with reasons)
unallowed_ipopt_options = {
    'wantsol': 'The solver interface requires the sol file to be created',
    'option_file_name': (
        'Pyomo generates the ipopt options file as part of the `solve` '
        'method.  Add all options to config.solver_options instead.'
    ),
}


@document_class_CONFIG(methods=['solve'])
class Ipopt(SolverBase):
    """Interface to the Ipopt NLP solver (NL file based)"""

    #: Global class configuration;
    #: see :ref:`pyomo.contrib.solver.solvers.ipopt.Ipopt::CONFIG`.
    CONFIG = IpoptConfig()

    #: cache of availability / version information
    _exe_cache: dict[str : tuple[int] | None] = {}

    #: default timeout to use when attempting to get the ipopt version number
    _version_timeout = 2

    def __init__(self, **kwds: Any) -> None:
        super().__init__(**kwds)

        #: Instance configuration;
        #: see :ref:`pyomo.contrib.solver.solvers.ipopt.Ipopt::CONFIG`.
        self.config = self.config

    def available(self) -> Availability:
        return (
            Availability.NotFound
            if self.version() is None
            else Availability.FullLicense
        )

    def version(self) -> tuple[int, int, int] | None:
        return self._get_version(self.config.executable.path())

    def _get_version(self, exe):
        try:
            return self._exe_cache[exe]
        except KeyError:
            pass
        if exe is None:
            # No executable (either we couldn't find a matching file, or
            # the file is not executable)
            self._exe_cache[None] = None
            return None
        # Run the executable and look for the version
        results = subprocess.run(
            [str(exe), '--version'],
            timeout=self._version_timeout,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            check=False,
        )
        # Note that we expect the command to run without error, AND that
        # it returns a string starting "ipopt <version>".  That prevents
        # us from trying to use other (even ASL) executables as if they
        # were ipopt
        fields = results.stdout.split(maxsplit=2)
        if results.returncode:
            ver = None
        elif len(fields) != 3 or fields[0].lower() != 'ipopt':
            ver = None
        else:
            try:
                ver = tuple(int(i) for i in fields[1].split('.'))
            except (ValueError, TypeError):
                ver = None
        if ver is None:
            logger.warning(
                f"Failed parsing Ipopt version: '{exe} --version':\n\n{results.stdout}"
            )
        self._exe_cache[exe] = ver
        return ver

    def has_linear_solver(self, linear_solver: str) -> bool:
        """Determine if Ipopt has access to the specified linear solver

        This solves a small problem to detect if the Ipopt executable
        has access to the specified linear solver.

        Parameters
        ----------
        linear_solver : str

            The linear solver to test.  Accepts any string that is valid
            for the ``linear_solver`` Ipopt option.

        """
        import pyomo.core as AML

        m = AML.ConcreteModel()
        m.x = AML.Var()
        m.o = AML.Objective(expr=(m.x - 2) ** 2)
        results = self.solve(
            m,
            tee=False,
            raise_exception_on_nonoptimal_result=False,
            load_solutions=False,
            solver_options={'linear_solver': linear_solver},
        )
        return 'running with linear solver' in results.solver_log

    def solve(self, model, **kwds) -> Results:
        "Solve a model using Ipopt"
        # Begin time tracking
        start_time = default_timer()
        # Allocate the results object so we can populate it as we go
        results = Results()
        results.timing_info.start_timestamp = datetime.datetime.now(
            datetime.timezone.utc
        )
        results.solver_name = self.name

        # Update configuration options, based on keywords passed to solve
        config: IpoptConfig = self.config(value=kwds, preserve_implicit=True)

        timer = config.timer
        if timer is None:
            timer = config.timer = HierarchicalTimer()

        # As we are about to run a solver, update the stale flag
        StaleFlagManager.mark_all_as_stale()

        with TempfileManager.new_context() as tempfile:
            if config.working_dir is None:
                dname = tempfile.mkdtemp()
            else:
                dname = config.working_dir
            if not os.path.exists(dname):
                os.mkdir(dname)
            # Because we are just "making up" a file name, it is better
            # to always generate a consistent and legal name, rather
            # than blindly follow what the user gave us.  We will use
            # `universal=True` here to make sure that double quotes are
            # translated, thereby guaranteeing that we should always
            # generate a legal base name (unless, of course, the user
            # put double quotes somewhere else in the path)
            basename = to_legal_filename(model.name, universal=True)
            nlfd, nl_fname = tempfile.mkstemp(
                suffix='.nl', prefix=basename, dir=dname, text=True, delete=False
            )
            results.extra_info.base_file_name = basename = nl_fname[:-3]
            for ext in ('.row', '.col', '.sol', '.opt'):
                if os.path.exists(basename + ext):
                    raise RuntimeError(
                        f"Solver interface file {basename + ext} already exists!"
                    )
            # Note: the ASL has an issue where string constants written
            # to the NL file (e.g. arguments in external functions) MUST
            # be terminated with '\n' regardless of platform.  We will
            # disable universal newlines in the NL file to prevent
            # Python from mapping those '\n' to '\r\n' on Windows.
            with (
                os.fdopen(nlfd, 'w', newline='\n', encoding='utf-8') as nl_file,
                open(basename + '.row', 'w', encoding='utf-8') as row_file,
                open(basename + '.col', 'w', encoding='utf-8') as col_file,
            ):
                timer.start('write_nl_file')
                try:
                    # Note: this is mapping the top-level
                    # symbolic_solver_labels onto the solver's writer
                    # config, and then that config is being used (in
                    # it's entirety) to set the NLWriter's CONFIG.
                    nl_info = NLWriter().write(
                        model,
                        nl_file,
                        row_file,
                        col_file,
                        config=config.writer_config,
                        symbolic_solver_labels=config.symbolic_solver_labels,
                    )
                    proven_infeasible = False
                except InfeasibleConstraintException:
                    proven_infeasible = True
                    nl_info = NLWriterInfo()
                timer.stop('write_nl_file')

            if proven_infeasible:
                results.termination_condition = TerminationCondition.provenInfeasible
                results.solution_status = SolutionStatus.noSolution
                results.extra_info.iteration_count = 0
            elif not nl_info.variables:
                if nl_info.eliminated_vars:
                    results.termination_condition = (
                        TerminationCondition.convergenceCriteriaSatisfied
                    )
                    results.solution_status = SolutionStatus.optimal
                    results.solution_loader = IpoptSolutionLoader(
                        sol_data=ASLSolFileData(), nl_info=nl_info
                    )
                else:
                    results.termination_condition = TerminationCondition.emptyModel
                    results.solution_status = SolutionStatus.noSolution
                results.extra_info.iteration_count = 0
            else:
                self._run_ipopt(results, config, nl_info, basename, timer)

        if (
            config.raise_exception_on_nonoptimal_result
            and results.solution_status != SolutionStatus.optimal
        ):
            raise NoOptimalSolutionError()

        if config.load_solutions:
            if results.solution_status == SolutionStatus.noSolution:
                raise NoSolutionError()
            results.solution_loader.load_vars()
            if (
                hasattr(model, 'dual')
                and isinstance(model.dual, Suffix)
                and model.dual.import_enabled()
            ):
                model.dual.update(results.solution_loader.get_duals())
            if (
                hasattr(model, 'rc')
                and isinstance(model.rc, Suffix)
                and model.rc.import_enabled()
            ):
                model.rc.update(results.solution_loader.get_reduced_costs())

        if (
            results.solution_status in {SolutionStatus.feasible, SolutionStatus.optimal}
            and len(nl_info.objectives) > 0
        ):
            if config.load_solutions:
                results.incumbent_objective = value(nl_info.objectives[0])
            else:
                results.incumbent_objective = value(
                    replace_expressions(
                        nl_info.objectives[0].expr,
                        substitution_map={
                            id(v): val
                            for v, val in results.solution_loader.get_primals().items()
                        },
                        descend_into_named_expressions=True,
                        remove_named_expressions=True,
                    )
                )

        results.solver_config = config

        # Capture/record end-time / wall-time
        results.timing_info.timer = timer
        results.timing_info.wall_time = default_timer() - start_time
        return results

    def _process_options(
        self, option_fname: str, options: dict[str, str | int | float]
    ) -> list[str]:
        # Look through the solver options and separate the command line
        # options from the options that must be sent via an options
        # file.  Raise an exception for any unallowable options.
        options_file_options = []
        cmd_line_options = []
        for key, val in options.items():
            if key in unallowed_ipopt_options:
                msg = unallowed_ipopt_options[key]
                raise ValueError(f"unallowed Ipopt option '{key}': {msg}")
            elif key in ipopt_command_line_options:
                cmd_line_options.append(_option_to_cmd(key, val))
            else:
                options_file_options.append(f"{key} {val}\n")
        # create the options file (if we need it)
        if options_file_options:
            with open(option_fname, 'w', encoding='utf-8') as OPT_FILE:
                OPT_FILE.writelines(options_file_options)
            cmd_line_options.append(_option_to_cmd('option_file_name', option_fname))
        # Return the (formatted) command line options
        return cmd_line_options

    def _run_ipopt(self, results, config, nl_info, basename, timer):
        # Get a copy of the environment to pass to the subprocess
        env = os.environ.copy()
        if nl_info.external_function_libraries:
            env['AMPLFUNC'] = amplfunc_merge(env, *nl_info.external_function_libraries)

        # Get the Ipopt executable and start building the command line
        exe = config.executable.path()
        if not exe:
            raise ApplicationError('ipopt executable not found')
        cmd = [exe, basename + '.nl', '-AMPL']

        # Process ipopt options (splitting them between command line
        # options and those that must be passed through the opt file)
        options = config.solver_options.value()
        # Map standard Pyomo solver options to Ipopt options: standard
        # options override ipopt-specific options.
        if config.threads and config.threads != 1:
            logger.log(
                logging.WARNING,
                msg=f"The `threads={config.threads}` option was specified, "
                f"but this is not used by {self.__class__.__name__}.",
            )
        if config.time_limit is not None:
            options['max_cpu_time'] = config.time_limit
        cmd.extend(self._process_options(basename + '.opt', options))

        results.solver_version = self._get_version(exe)
        results.extra_info.add(
            'command_line', ConfigValue(cmd, visibility=ADVANCED_OPTION)
        )

        # This seems silly, but we have to give the subprocess slightly
        # longer to finish than ipopt, otherwise we may kill the
        # subprocess before ipopt has a chance to write the SOL file.
        # We will add 1% (with a min of 1 second and max of 100 seconds).
        timeout = config.time_limit
        if timeout is not None:
            timeout = timeout + min(max(1.0, 0.01 * timeout), 100.0)

        # Call ipopt - passing the files via the subprocess
        ostreams = [io.StringIO()] + config.tee
        timer.start('subprocess')
        try:
            with TeeStream(*ostreams) as t:
                process = subprocess.run(
                    cmd,
                    timeout=timeout,
                    env=env,
                    universal_newlines=True,
                    stdout=t.STDOUT,
                    stderr=t.STDERR,
                    check=False,
                )
        except OSError:
            err = sys.exc_info()[1]
            msg = 'Could not execute the command: %s\tError message: %s'
            raise ApplicationError(msg % (cmd, err))
        finally:
            timer.stop('subprocess')

        results.solver_log = ostreams[0].getvalue()
        results.extra_info.return_code = process.returncode
        if process.returncode:
            results.termination_condition = TerminationCondition.error

        # This is the data we need to parse to get the iterations
        # and time
        timer.start('parse_log')
        parsed_output_data = self._parse_ipopt_output(results.solver_log)
        results.extra_info.iteration_count = parsed_output_data.pop('iters', None)
        _timing = parsed_output_data.pop('cpu_seconds', None)
        if _timing:
            results.timing_info.update(_timing)
        # Save the iteration log, but mark it as an "advanced" result
        iter_log = parsed_output_data.pop('iteration_log', None)
        if iter_log is not None:
            results.extra_info.add(
                'iteration_log', ConfigList(iter_log, visibility=ADVANCED_OPTION)
            )
        results.extra_info.update(parsed_output_data)
        timer.stop('parse_log')

        timer.start('parse_sol')
        if os.path.isfile(basename + '.sol'):
            with open(basename + '.sol', 'r', encoding='utf-8') as sol_file:
                sol_data = parse_asl_sol_file(sol_file)
        else:
            sol_data = ASLSolFileData()
        results.solution_loader = IpoptSolutionLoader(
            sol_data=sol_data, nl_info=nl_info
        )
        timer.stop('parse_sol')

        # Initialize the solver message, solution loader solution
        # status and termination condition:
        asl_solve_code_to_solution_status(sol_data, results)

    def _parse_ipopt_output(self, output: str) -> Dict[str, Any]:
        parsed_data = {}

        # Stop parsing if there is nothing to parse
        if not output:
            logger.log(
                logging.WARNING,
                "Returned output from ipopt was empty. Cannot parse for additional data.",
            )
            return parsed_data

        # Extract number of iterations
        iter_match = re.search(r'Number of Iterations.*:\s+(\d+)', output)
        if iter_match:
            parsed_data['iters'] = int(iter_match.group(1))
        # Gather all the iteration data
        iter_table = re.findall(r'^(?:\s*\d+.*?)$', output, re.MULTILINE)
        if iter_table:
            columns = [
                ("iter", int),
                ("objective", float),
                ("inf_pr", float),
                ("inf_du", float),
                ("lg_mu", float),
                ("d_norm", float),
                ("lg_rg", float),
                ("alpha_du", float),
                ("alpha_pr", float),
                ("ls", int),
            ]
            iterations = []
            n_expected_columns = len(columns)
            iter_idx = columns.index(('iter', int))
            alpha_pr_idx = columns.index(('alpha_pr', float))

            for line in iter_table:
                tokens = line.strip().split()
                # IPOPT sometimes mashes the first two column values together
                # (e.g., "2r-4.93e-03"). We need to split them.
                if '-' in tokens[iter_idx]:
                    # This happens rarely, so we are OK with this
                    # portion of the parser being a little less
                    # efficient (e.g., reallocating the tokens list, and
                    # performing index math)
                    tkn = tokens[iter_idx]
                    idx = tkn.index('-')
                    tokens[iter_idx : iter_idx + 1] = tkn[:idx], tkn[idx:]

                # Extract restoration flag from 'iter'
                restoration = tokens[iter_idx].endswith("r")
                if restoration:
                    tokens[iter_idx] = tokens[iter_idx][:-1]

                # Separate alpha_pr into numeric part and optional tag (f, D, R, etc.)
                step_acceptance = tokens[alpha_pr_idx][-1]
                if step_acceptance in _ALPHA_PR_CHARS:
                    tokens[alpha_pr_idx] = tokens[alpha_pr_idx][:-1]
                else:
                    step_acceptance = None

                try:
                    iter_data = {
                        key: None if t == '-' else cast(t)
                        for (key, cast), t in zip(columns, tokens)
                    }
                except (ValueError, TypeError):
                    logger.error(
                        "Error parsing Ipopt log entry:\n"
                        f"\t{sys.exc_info()[1]}\n\t{line}"
                    )
                    # Fall-back on a simpler (but slower) parse: extract
                    # the fields, and cast to float what we can.  The
                    # point here is the parser should never fail with an
                    # exception (even if it fails to parse some of the
                    # log)
                    iter_data = {}
                    for (key, cast), t in zip(columns, tokens):
                        if t == '-':
                            t = None
                        else:
                            try:
                                t = cast(t)
                            except:
                                pass
                        iter_data[key] = t

                iter_data["restoration"] = restoration
                iter_data["step_acceptance"] = step_acceptance

                # Capture optional IPOPT diagnostic tags if present
                if len(tokens) > n_expected_columns:
                    iter_data['diagnostic_tags'] = " ".join(tokens[n_expected_columns:])

                iterations.append(iter_data)

            parsed_data['iteration_log'] = iterations

            if len(iterations) != parsed_data.get('iters', 0) + 1:
                n_iter = parsed_data.get('iters', 0)
                logger.warning(
                    f"Total number of iteration records parsed {len(iterations)} does "
                    f"not match the number of iterations ({n_iter}) plus one."
                )

        # Extract scaled and unscaled table
        scaled_unscaled_match = re.search(
            r'''
            Objective\.*:\s*([-+eE0-9.]+)\s+([-+eE0-9.]+)\s*
            Dual\ infeasibility\.*:\s*([-+eE0-9.]+)\s+([-+eE0-9.]+)\s*
            Constraint\ violation\.*:\s*([-+eE0-9.]+)\s+([-+eE0-9.]+)\s*
            (?:Variable\ bound\ violation:\s*([-+eE0-9.]+)\s+([-+eE0-9.]+)\s*)?
            Complementarity\.*:\s*([-+eE0-9.]+)\s+([-+eE0-9.]+)\s*
            Overall\ NLP\ error\.*:\s*([-+eE0-9.]+)\s+([-+eE0-9.]+)
            ''',
            output,
            re.DOTALL | re.VERBOSE,
        )

        if scaled_unscaled_match:
            groups = scaled_unscaled_match.groups()
            all_fields = [
                "incumbent_objective",
                "dual_infeasibility",
                "constraint_violation",
                "variable_bound_violation",  # optional
                "complementarity_error",
                "overall_nlp_error",
            ]
            # Filter out None values and create final fields and values.
            # Nones occur in old-style IPOPT output (<= 3.13)
            zipped = [
                (field, scaled, unscaled)
                for field, scaled, unscaled in zip(
                    all_fields, groups[0::2], groups[1::2]
                )
                if scaled is not None and unscaled is not None
            ]
            scaled = {k: float(s) for k, s, _ in zipped}
            unscaled = {k: float(u) for k, _, u in zipped}
            parsed_data.update(unscaled)
            parsed_data['final_scaled_results'] = scaled

        # Newer versions of IPOPT no longer separate timing into
        # two different values. This is so we have compatibility with
        # both new and old versions
        parsed_data['cpu_seconds'] = {
            k.strip(): float(v)
            for k, v in re.findall(
                r'Total(?: CPU)? sec(?:ond)?s in ([^=]+)=\s*([0-9.]+)', output
            )
        }

        return parsed_data


class LegacyIpoptSolver(LegacySolverWrapper, Ipopt):
    def _process_options(
        self, option_fname: str, options: dict[str, str | int | float]
    ) -> list[str]:
        # The old Ipopt solver would map solver_options starting with
        # "OF_" to the options file.  That is no longer needed, so we
        # will strip off any "OF_" that we find
        for opt in list(options):
            if opt.startswith('OF_'):
                options[opt[3:]] = options.pop(opt)
        return super()._process_options(option_fname, options)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from pyomo.common.log import LoggingIntercept
import pyomo.common.unittest as unittest
from pyomo.contrib.solver.common.factory import SolverFactory

from pyomo.environ import (
    ConcreteModel,
    Var,
    Constraint,
    value,
    Binary,
    NonNegativeReals,
    Objective,
    Set,
)

gurobi_direct = SolverFactory('gurobi_direct')
gurobi_direct_minlp = SolverFactory('gurobi_direct_minlp')
gurobi_persistent = SolverFactory('gurobi_persistent')


class TestGurobiWarmStart(unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()
        m.S = Set(initialize=[1, 2, 3, 4, 5])
        m.y = Var(m.S, domain=Binary)
        m.x = Var(m.S, domain=NonNegativeReals)
        m.obj = Objective(expr=sum(m.x[i] for i in m.S))

        @m.Constraint(m.S)
        def cons(m, i):
            if i % 2 == 0:
                return m.x[i] + i * m.y[i] >= 3 * i
            else:
                return m.x[i] - i * m.y[i] >= 3 * i

        # define a suboptimal MIP start
        for i in m.S:
            m.y[i] = 1
        # objective will be 4 + 4 + 12 + 8 + 20 = 48

        return m

    def check_optimal_soln(self, m):
        # check that we got the optimal solution:
        # y[1] = 0, x[1] = 3
        # y[2] = 1, x[2] = 4
        # y[3] = 0, x[3] = 9
        # y[4] = 1, x[4] = 8
        # y[5] = 0, x[5] = 15
        x = {1: 3, 2: 4, 3: 9, 4: 8, 5: 15}
        self.assertEqual(value(m.obj), 39)
        for i in m.S:
            if i % 2 == 0:
                self.assertEqual(value(m.y[i]), 1)
            else:
                self.assertEqual(value(m.y[i]), 0)
            self.assertEqual(value(m.x[i]), x[i])

    @unittest.skipUnless(gurobi_direct.available(), "needs Gurobi Direct interface")
    @unittest.pytest.mark.solver("gurobi_direct")
    def test_gurobi_direct_warm_start(self):
        m = self.make_model()

        gurobi_direct.config.warmstart_discrete_vars = True
        logger = logging.getLogger('tee')
        with LoggingIntercept(module='tee', level=logging.INFO) as LOG:
            gurobi_direct.solve(m, tee=logger)
        self.assertIn(
            "User MIP start produced solution with objective 48", LOG.getvalue()
        )
        self.check_optimal_soln(m)

    @unittest.skipUnless(
        gurobi_direct_minlp.available(), "needs Gurobi Direct MINLP interface"
    )
    @unittest.pytest.mark.solver("gurobi_direct_minlp")
    def test_gurobi_minlp_warmstart(self):
        m = self.make_model()

        gurobi_direct_minlp.config.warmstart_discrete_vars = True
        logger = logging.getLogger('tee')
        with LoggingIntercept(module='tee', level=logging.INFO) as LOG:
            gurobi_direct_minlp.solve(m, tee=logger)
        self.assertIn(
            "User MIP start produced solution with objective 48", LOG.getvalue()
        )
        self.check_optimal_soln(m)

    @unittest.skipUnless(
        gurobi_persistent.available(), "needs Gurobi persistent interface"
    )
    @unittest.pytest.mark.solver("gurobi_persistent")
    def test_gurobi_persistent_warmstart(self):
        m = self.make_model()

        gurobi_persistent.config.warmstart_discrete_vars = True
        gurobi_persistent.set_instance(m)
        logger = logging.getLogger('tee')
        with LoggingIntercept(module='tee', level=logging.INFO) as LOG:
            gurobi_persistent.solve(m, tee=logger)
        self.assertIn(
            "User MIP start produced solution with objective 48", LOG.getvalue()
        )
        self.check_optimal_soln(m)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os

import pyomo.core.expr as EXPR


def compare_floats(base, test, abstol=1e-14, reltol=1e-14):
    base = base.split()
    test = test.split()
    if len(base) != len(test):
        return False
    for i, b in enumerate(base):
        if b.strip() == test[i].strip():
            continue
        try:
            b = float(b)
            t = float(test[i])
        except:
            return False
        if abs(b - t) < abstol:
            continue
        if abs((b - t) / max(abs(b), abs(t))) < reltol:
            continue
        return False
    return True


def load_baseline(baseline, testfile, extension, version):
    with open(testfile, 'r') as FILE:
        test = FILE.read()
    if baseline.endswith(f'.{extension}'):
        _tmp = [baseline[:-3]]
    else:
        _tmp = baseline.split(f'.{extension}.', 1)
    _tmp.insert(1, f'expr{int(EXPR.Mode.CURRENT)}')
    _tmp.insert(2, version)
    if not os.path.exists('.'.join(_tmp)):
        _tmp.pop(1)
        if not os.path.exists('.'.join(_tmp)):
            _tmp = []
    if _tmp:
        baseline = '.'.join(_tmp)
    with open(baseline, 'r') as FILE:
        base = FILE.read()
    return base, test, baseline, testfile



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base.PyomoModel import Model
from pyomo.core.base.block import Block, BlockData
from pyomo.core.kernel.block import IBlock
from pyomo.opt.base.solvers import OptSolver
from pyomo.core.base import SymbolMap, NumericLabeler, TextLabeler
import pyomo.common
from pyomo.common.errors import ApplicationError
from pyomo.common.collections import ComponentMap, ComponentSet, Bunch
from pyomo.common.tempfiles import TempfileManager
import pyomo.opt.base.solvers
from pyomo.opt.base.formats import ResultsFormat
from pyomo.core.staleflag import StaleFlagManager


class DirectOrPersistentSolver(OptSolver):
    """
    This is a base class for both direct and persistent solvers. Direct solver interfaces do not use any file io.
    Rather, they interface directly with the python bindings for the specific solver. Persistent solver interfaces
    are similar except that they "remember" their model. Thus, persistent solver interfaces allow incremental changes
    to the solver model (e.g., the gurobi python model or the cplex python model). Note that users are responsible
    for notifying the persistent solver interfaces when changes are made to the corresponding pyomo model.

    Parameters
    ----------
    type: str
        String indicating the class type of the solver instance.
    name: str
        String representing either the class type of the solver instance or an assigned name.
    doc: str
        Documentation for the solver
    options: dict
        Dictionary of solver options
    """

    def __init__(self, **kwds):
        OptSolver.__init__(self, **kwds)

        self._pyomo_model = None
        """The pyomo model being solved."""

        self._solver_model = None
        """The python instance of the solver model (e.g., the gurobipy Model instance)."""

        self._symbol_map = SymbolMap()
        """A symbol map used to map between pyomo components and their names used with the solver."""

        self._labeler = None
        """The labeler for creating names for the solver model components."""

        self._pyomo_var_to_solver_var_map = ComponentMap()
        self._solver_var_to_pyomo_var_map = dict()
        """A dictionary mapping pyomo Var's to the solver variables."""

        self._pyomo_con_to_solver_con_map = dict()
        self._solver_con_to_pyomo_con_map = dict()
        """A dictionary mapping pyomo constraints to solver constraints."""

        self._vars_referenced_by_con = ComponentMap()
        """A dictionary mapping constraints to a ComponentSet containing the pyomo variables referenced by that
        constraint. This is primarily needed for the persistent solvers. When a constraint is deleted, we need
        to decrement the number of times those variables are referenced (see self._referenced_variables)."""

        self._vars_referenced_by_obj = ComponentSet()
        """A set containing the pyomo variables referenced by that the objective.
        This is primarily needed for the persistent solvers. When a the objective is deleted, we need
        to decrement the number of times those variables are referenced (see self._referenced_variables)."""

        self._objective = None
        """The pyomo Objective object currently being used with the solver."""

        self.results = None
        """A results object return from the solve method."""

        self._skip_trivial_constraints = False
        """A bool. If True, then any constraints with a constant body will not be added to the solver model.
        Be careful with this. If a trivial constraint is skipped then that constraint cannot be removed from
        a persistent solver (an error will be raised if a user tries to remove a non-existent constraint)."""

        self._output_fixed_variable_bounds = False
        """A bool. If False then an error will be raised if a fixed variable is used in one of the solver constraints.
        This is useful for catching bugs. Ordinarily a fixed variable should appear as a constant value in the
        solver constraints. If True, then the error will not be raised."""

        self._python_api_exists = False
        """A bool indicating whether or not the python api is available for the specified solver."""

        self._version = None
        """The version of the solver."""

        self._symbolic_solver_labels = False
        """A bool. If true then the solver components will be given names corresponding to the pyomo component names."""

        self._capabilities = Bunch()

        self._referenced_variables = ComponentMap()
        """dict: {var: count} where count is the number of constraints/objective referencing the var"""

        self._keepfiles = False
        """A bool. If True, then the solver log will be saved."""

        self._save_results = True
        """A bool. This is used for backwards compatibility. If True, the solution will be loaded into the Solution
        object that gets placed on the SolverResults object. This way, users can do model.solutions.load_from(results)
        to load solutions into their model. However, it is more efficient to bypass the Solution object and load
        the results directly from the solver object. If False, the solution will not be loaded into the Solution
        object."""

    def _presolve(self, **kwds):
        warmstart_flag = kwds.pop('warmstart', False)
        self._keepfiles = kwds.pop('keepfiles', False)
        self._save_results = kwds.pop('save_results', True)

        # create a context in the temporary file manager for
        # this plugin - is "pop"ed in the _postsolve method.
        TempfileManager.push()

        self.results = None

        model = self._pyomo_model

        # this implies we have a custom solution "parser",
        # preventing the OptSolver _presolve method from
        # creating one
        self._results_format = ResultsFormat.soln
        # use the base class _presolve to consume the
        # important keywords
        OptSolver._presolve(self, **kwds)

        # ***********************************************************
        # The following code is only needed for backwards compatibility of load_solutions=False.
        # If we ever only want to support the load_vars, load_duals, etc. methods, then this can be deleted.
        if self._save_results:
            self._smap_id = id(self._symbol_map)
            if isinstance(self._pyomo_model, IBlock):
                # BIG HACK (see pyomo.core.kernel write function)
                if not hasattr(self._pyomo_model, "._symbol_maps"):
                    setattr(self._pyomo_model, "._symbol_maps", {})
                getattr(self._pyomo_model, "._symbol_maps")[
                    self._smap_id
                ] = self._symbol_map
            else:
                self._pyomo_model.solutions.add_symbol_map(self._symbol_map)
        # ***********************************************************

        if warmstart_flag:
            if self.warm_start_capable():
                self._warm_start()
            else:
                raise ValueError(
                    '{0} solver plugin is not capable of warmstart.'.format(type(self))
                )

        if self._log_file is None:
            self._log_file = TempfileManager.create_tempfile(suffix='.log')

    """ This method should be implemented by subclasses."""

    def _apply_solver(self):
        raise NotImplementedError('This method should be implemented by subclasses')

    """ This method should be implemented by subclasses."""

    def _postsolve(self):
        return OptSolver._postsolve(self)

    """ This method should be implemented by subclasses."""

    def _set_instance(self, model, kwds={}):
        if not isinstance(model, (Model, IBlock, Block, BlockData)):
            msg = (
                "The problem instance supplied to the {0} plugin "
                "'_presolve' method must be a Model or a Block".format(type(self))
            )
            raise ValueError(msg)
        self._pyomo_model = model
        self._symbolic_solver_labels = kwds.pop(
            'symbolic_solver_labels', self._symbolic_solver_labels
        )
        self._skip_trivial_constraints = kwds.pop(
            'skip_trivial_constraints', self._skip_trivial_constraints
        )
        self._output_fixed_variable_bounds = kwds.pop(
            'output_fixed_variable_bounds', self._output_fixed_variable_bounds
        )
        self._pyomo_var_to_solver_var_map = ComponentMap()
        self._solver_var_to_pyomo_var_map = dict()
        self._pyomo_con_to_solver_con_map = dict()
        self._solver_con_to_pyomo_con_map = dict()
        self._vars_referenced_by_con = ComponentMap()
        self._vars_referenced_by_obj = ComponentSet()
        self._referenced_variables = ComponentMap()
        self._objective_label = None
        self._objective = None

        self._symbol_map = SymbolMap()

        if self._symbolic_solver_labels:
            self._labeler = TextLabeler()
        else:
            self._labeler = NumericLabeler('x')

    def _add_block(self, block):
        for var in block.component_data_objects(
            ctype=pyomo.core.base.var.Var, descend_into=True, active=True, sort=True
        ):
            self._add_var(var)

        for sub_block in block.block_data_objects(descend_into=True, active=True):
            for con in sub_block.component_data_objects(
                ctype=pyomo.core.base.constraint.Constraint,
                descend_into=False,
                active=True,
                sort=True,
            ):
                if (not con.has_lb()) and (not con.has_ub()):
                    assert not con.equality
                    continue  # non-binding, so skip
                self._add_constraint(con)

            for con in sub_block.component_data_objects(
                ctype=pyomo.core.base.sos.SOSConstraint,
                descend_into=False,
                active=True,
                sort=True,
            ):
                self._add_sos_constraint(con)

            obj_counter = 0
            for obj in sub_block.component_data_objects(
                ctype=pyomo.core.base.objective.Objective,
                descend_into=False,
                active=True,
            ):
                obj_counter += 1
                if obj_counter > 1:
                    raise ValueError(
                        "Solver interface does not support multiple objectives."
                    )
                self._set_objective(obj)

    """ This method should be implemented by subclasses."""

    def _set_objective(self, obj):
        raise NotImplementedError("This method should be implemented by subclasses")

    """ This method should be implemented by subclasses."""

    def _add_constraint(self, con):
        raise NotImplementedError("This method should be implemented by subclasses")

    """ This method should be implemented by subclasses."""

    def _add_sos_constraint(self, con):
        raise NotImplementedError("This method should be implemented by subclasses")

    """ This method should be implemented by subclasses."""

    def _add_var(self, var):
        raise NotImplementedError("This method should be implemented by subclasses")

    """ This method should be implemented by subclasses."""

    def _get_expr_from_pyomo_repn(self, repn, max_degree=None):
        raise NotImplementedError("This method should be implemented by subclasses")

    """ This method should be implemented by subclasses."""

    def _get_expr_from_pyomo_expr(self, expr, max_degree=None):
        raise NotImplementedError("This method should be implemented by subclasses")

    """ This method should be implemented by subclasses."""

    def _load_vars(self, vars_to_load):
        raise NotImplementedError("This method should be implemented by subclasses")

    def load_vars(self, vars_to_load=None):
        """
        Load the values from the solver's variables into the corresponding pyomo variables.

        Parameters
        ----------
        vars_to_load: list of Var
        """
        self._load_vars(vars_to_load)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    """ This method should be implemented by subclasses."""

    def warm_start_capable(self):
        raise NotImplementedError('This method should be implemented by subclasses')

    def _warm_start(self):
        raise NotImplementedError(
            'If a subclass can warmstart, then it should implement this method.'
        )

    def available(self, exception_flag=True):
        """True if the solver is available."""

        _api = getattr(self, '_python_api_exists', False)
        if exception_flag and not _api:
            raise ApplicationError(
                "No Python bindings available for %s solver plugin" % (type(self),)
            )
        return bool(_api)

    def _get_version(self):
        if self._version is None:
            return pyomo.opt.base.solvers._extract_version('')
        return self._version



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests for the MindtPy solver."""

from pyomo.core.expr.calculus.diff_with_sympy import differentiate_available
import pyomo.common.unittest as unittest
from pyomo.environ import SolverFactory, value, maximize
from pyomo.opt import TerminationCondition
from pyomo.common.dependencies import numpy_available, scipy_available
from pyomo.contrib.mindtpy.tests.MINLP_simple import SimpleMINLP as SimpleMINLP

model_list = [SimpleMINLP(grey_box=True)]

if SolverFactory('appsi_highs').available(exception_flag=False) and SolverFactory(
    'appsi_highs'
).version() >= (1, 7, 0):
    required_solvers = ('cyipopt', 'appsi_highs')
else:
    required_solvers = ('cyipopt', 'glpk')

if all(SolverFactory(s).available(exception_flag=False) for s in required_solvers):
    subsolvers_available = True
else:
    subsolvers_available = False


@unittest.skipIf(model_list[0] is None, 'Unable to generate the Grey Box model.')
@unittest.skipIf(
    not subsolvers_available,
    'Required subsolvers %s are not available' % (required_solvers,),
)
@unittest.skipIf(
    not differentiate_available, 'Symbolic differentiation is not available'
)
class TestMindtPy(unittest.TestCase):
    """Tests for the MindtPy solver plugin."""

    def check_optimal_solution(self, model, places=1):
        for var in model.optimal_solution:
            self.assertAlmostEqual(
                var.value, model.optimal_solution[var], places=places
            )

    def test_OA_rNLP(self):
        """Test the outer approximation decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='OA',
                    init_strategy='rNLP',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    calculate_dual_at_solution=True,
                    nlp_solver_args={
                        'options': {
                            'hessian_approximation': 'limited-memory',
                            'linear_solver': 'mumps',
                        }
                    },
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.deprecation import deprecated
from pyomo.common.modeling import NOTSET
import pyomo.core.expr as EXPR
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.kernel.base import ICategorizedObject, _abstract_readwrite_property
from pyomo.core.kernel.container_utils import define_simple_containers
from pyomo.core.expr.numvalue import (
    NumericValue,
    is_fixed,
    is_constant,
    is_potentially_variable,
    is_numeric_data,
    value,
)


class IIdentityExpression(NumericValue):
    """The interface for classes that simply wrap another
    expression and perform no additional operations.

    Derived classes should declare an _expr attribute or
    override all implemented methods.
    """

    __slots__ = ()

    PRECEDENCE = 0

    ASSOCIATIVITY = EXPR.OperatorAssociativity.NON_ASSOCIATIVE

    @property
    def expr(self):
        return self._expr

    #
    # Implement the NumericValue abstract methods
    #

    def __call__(self, exception=NOTSET):
        """Compute the value of this expression.

        Args:
            exception (bool): Indicates if an exception
                should be raised when instances of
                NumericValue fail to evaluate due to one or
                more objects not being initialized to a
                numeric value (e.g, one or more variables in
                an algebraic expression having the value
                None). Default is :const:`True`.

        Returns:
            numeric value or None
        """
        exception = _type_check_exception_arg(self, exception)
        return value(self._expr, exception=exception)

    def is_fixed(self):
        """A boolean indicating whether this expression is fixed."""
        return is_fixed(self._expr)

    def is_parameter_type(self):
        """A boolean indicating whether this expression is a parameter object."""
        return False

    def is_variable_type(self):
        """A boolean indicating whether this expression is a
        variable object."""
        return False

    def is_named_expression_type(self):
        """A boolean indicating whether this in a named expression."""
        return True

    def is_expression_type(self, expression_system=None):
        """A boolean indicating whether this in an expression."""
        return True

    @property
    def _args_(self):
        """A tuple of subexpressions involved in this expressions operation."""
        return (self._expr,)

    @property
    def args(self):
        """A tuple of subexpressions involved in this expressions operation."""
        return (self._expr,)

    def nargs(self):
        """Length of self._nargs()"""
        return 1

    def arg(self, i):
        if i != 0:
            raise KeyError("Unexpected argument id")
        return self._expr

    def polynomial_degree(self):
        """The polynomial degree of the stored expression."""
        if self.is_fixed():
            return 0
        return self._expr.polynomial_degree()

    def _compute_polynomial_degree(self, values):
        return values[0]

    def to_string(self, verbose=None, labeler=None, smap=None, compute_values=False):
        """Convert this expression into a string."""
        return EXPR.expression_to_string(
            self,
            verbose=verbose,
            labeler=labeler,
            smap=smap,
            compute_values=compute_values,
        )

    def _to_string(self, values, verbose, smap):
        if verbose:
            name = self.getname()
            if name == None:
                return "<%s>{%s}" % (self.__class__.__name__, values[0])
            else:
                if name[0] == '<':
                    name = ""
                return "%s{%s}" % (name, values[0])
        if self._expr is None:
            return "%s{Undefined}" % str(self)
        return values[0]

    def _apply_operation(self, result):
        return result[0]

    def _is_fixed(self, values):
        return values[0]

    def create_node_with_local_data(self, values):
        """
        Construct an expression after constructing the
        contained expression.

        This class provides a consistent interface for constructing a
        node, which is used in tree visitor scripts.
        """
        return self.__class__(expr=values[0])

    def is_constant(self):
        raise NotImplementedError  # pragma:nocover

    def is_potentially_variable(self):
        raise NotImplementedError  # pragma:nocover

    def clone(self):
        raise NotImplementedError  # pragma:nocover


@deprecated(
    "noclone() is deprecated and can be omitted: "
    "Pyomo expressions natively support shared subexpressions.",
    version='6.6.2',
)
def noclone(expr):
    try:
        if expr.is_potentially_variable():
            return expression(expr)
    except AttributeError:
        pass
    try:
        if is_constant(expr):
            return expr
    except:
        return expr
    return data_expression(expr)


class IExpression(ICategorizedObject, IIdentityExpression):
    """
    The interface for mutable expressions.
    """

    __slots__ = ()

    #
    # Implementations can choose to define these
    # properties as using __slots__, __dict__, or
    # by overriding the @property method
    #

    expr = _abstract_readwrite_property(doc="The stored expression")

    #
    # Override some of the NumericValue methods implemented
    # by the base class
    #

    def is_constant(self):
        """A boolean indicating whether this expression is constant."""
        return False

    def is_potentially_variable(self):
        """A boolean indicating whether this expression can
        reference variables."""
        return True

    def clone(self):
        """Return a clone of this expression (no-op)."""
        return self


class expression(IExpression):
    """A named, mutable expression."""

    _ctype = IExpression
    __slots__ = ("_parent", "_storage_key", "_active", "_expr", "__weakref__")

    def __init__(self, expr=None):
        self._parent = None
        self._storage_key = None
        self._active = True
        self._expr = None

        # call the setter
        self.expr = expr

    #
    # Define the IExpression abstract methods
    #

    @property
    def expr(self):
        return self._expr

    @expr.setter
    def expr(self, expr):
        self._expr = expr


class data_expression(expression):
    """A named, mutable expression that is restricted to
    storage of data expressions. An exception will be raised
    if an expression is assigned that references (or is
    allowed to reference) variables."""

    __slots__ = ()

    #
    # Overload a few methods
    #

    def is_potentially_variable(self):
        """A boolean indicating whether this expression can
        reference variables."""
        return False

    def polynomial_degree(self):
        """Always return zero because we always validate
        that the stored expression can never reference
        variables."""
        return 0

    @property
    def expr(self):
        return self._expr

    @expr.setter
    def expr(self, expr):
        if (expr is not None) and (not is_numeric_data(expr)):
            raise ValueError("Expression is not restricted to numeric data.")
        self._expr = expr


# inserts class definitions for simple _tuple, _list, and
# _dict containers into this module
define_simple_containers(globals(), "expression", IExpression)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr import numvalue, numeric_expr, boolean_value, logical_expr
from pyomo.core.expr.numvalue import (
    value,
    is_constant,
    is_fixed,
    is_variable_type,
    is_potentially_variable,
    NumericValue,
    ZeroConstant,
    native_numeric_types,
    native_types,
    polynomial_degree,
)
from pyomo.core.expr.boolean_value import BooleanValue
from pyomo.core.expr import (
    log,
    log10,
    sin,
    cos,
    tan,
    cosh,
    sinh,
    tanh,
    asin,
    acos,
    atan,
    exp,
    sqrt,
    asinh,
    acosh,
    atanh,
    ceil,
    floor,
    Expr_if,
    inequality,
    linear_expression,
    nonlinear_expression,
    land,
    lor,
    equivalent,
    exactly,
    atleast,
    atmost,
    implies,
    lnot,
    xor,
)
from pyomo.core.expr.calculus.derivatives import differentiate
from pyomo.core.expr.taylor_series import taylor_series_expansion

from pyomo.core.kernel import (
    base,
    homogeneous_container,
    heterogeneous_container,
    variable,
    constraint,
    matrix_constraint,
    parameter,
    expression,
    objective,
    sos,
    suffix,
    block,
    piecewise_library,
    set_types,
)

#
# declare deprecation paths for removed modules and attributes
#
from pyomo.common.deprecation import relocated_module_attribute, moved_module

relocated_module_attribute(
    'component_map',
    'pyomo.common.collections.component_map',
    msg='The pyomo.core.kernel.component_map module is deprecated.  '
    'Import ComponentMap from pyomo.common.collections.',
    version='5.7.1',
    f_globals=globals(),
)
relocated_module_attribute(
    'component_set',
    'pyomo.common.collections.component_set',
    msg='The pyomo.core.kernel.component_map module is deprecated.  '
    'Import ComponentMap from pyomo.common.collections.',
    version='5.7.1',
    f_globals=globals(),
)

moved_module(
    "pyomo.core.kernel.component_map",
    "pyomo._archive.component_map",
    msg='The pyomo.core.kernel.component_map module is deprecated.  '
    'Import ComponentMap from pyomo.common.collections.',
    version='5.7.1',
)
moved_module(
    "pyomo.core.kernel.component_set",
    "pyomo._archive.component_set",
    msg='The pyomo.core.kernel.component_set module is deprecated.  '
    'Import ComponentSet from pyomo.common.collections.',
    version='5.7.1',
)
moved_module(
    "pyomo.core.kernel.register_numpy_types",
    "pyomo._archive.register_numpy_types",
    msg="pyomo.core.kernel.register_numpy_types is deprecated.  NumPy type "
    "registration is handled automatically by pyomo.common.dependencies.numpy",
    version='6.1',
)
del relocated_module_attribute, moved_module



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.fileutils import find_library
from pyomo.contrib.pynumero.linalg.utils import validate_index, validate_value, _NotSet
import numpy.ctypeslib as npct
import numpy as np
import ctypes
import os


class MA57Interface:
    libname = _NotSet

    @classmethod
    def available(cls):
        if cls.libname is _NotSet:
            cls.libname = find_library('pynumero_MA57')
        if cls.libname is None:
            return False
        return os.path.exists(cls.libname)

    def __init__(self, work_factor=None, fact_factor=None, ifact_factor=None):
        if not MA57Interface.available():
            raise RuntimeError('Could not find pynumero_MA57 library.')

        self.work_factor = work_factor
        self.fact_factor = fact_factor
        self.ifact_factor = ifact_factor

        self.lib = ctypes.cdll.LoadLibrary(self.libname)

        array_1d_double = npct.ndpointer(dtype=np.double, ndim=1, flags='CONTIGUOUS')
        array_2d_double = npct.ndpointer(dtype=np.double, ndim=2, flags='CONTIGUOUS')
        array_1d_int = npct.ndpointer(dtype=np.intc, ndim=1, flags='CONTIGUOUS')

        # Declare arg and res types of functions:

        # Do I need to specify that this function takes no argument?
        self.lib.new_MA57_struct.restype = ctypes.c_void_p
        # return type is pointer to MA57_struct. Why do I use c_void_p here?

        self.lib.free_MA57_struct.argtypes = [ctypes.c_void_p]

        self.lib.set_icntl.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
        # Do I need to specify that this function returns nothing?
        self.lib.get_icntl.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.get_icntl.restype = ctypes.c_int

        self.lib.set_cntl.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_double]
        self.lib.get_cntl.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.get_cntl.restype = ctypes.c_double

        self.lib.get_info.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.get_info.restype = ctypes.c_int

        self.lib.get_rinfo.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.get_rinfo.restype = ctypes.c_double

        self.lib.alloc_keep.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.alloc_work.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.alloc_fact.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.alloc_ifact.argtypes = [ctypes.c_void_p, ctypes.c_int]

        self.lib.set_nrhs.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.set_lrhs.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.set_job.argtypes = [ctypes.c_void_p, ctypes.c_int]

        self.lib.do_symbolic_factorization.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_int,
            array_1d_int,
            array_1d_int,
        ]
        self.lib.do_numeric_factorization.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_int,
            array_1d_double,
        ]
        self.lib.do_backsolve.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            array_2d_double,
        ]
        self.lib.do_iterative_refinement.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_int,
            array_1d_double,
            array_1d_int,
            array_1d_int,
            array_1d_double,
            array_1d_double,
            array_1d_double,
        ]
        self.lib.do_reallocation.argtypes = [
            ctypes.c_void_p,
            ctypes.c_int,
            ctypes.c_double,
            ctypes.c_int,
        ]

        self.icntl_len = 20
        self.cntl_len = 5
        self.info_len = 40
        self.rinfo_len = 20

        self._ma57 = self.lib.new_MA57_struct()

    def __del__(self):
        self.lib.free_MA57_struct(self._ma57)

    def set_icntl(self, i, val):
        validate_index(i, self.icntl_len, 'ICNTL')
        validate_value(i, int, 'ICNTL')
        # NOTE: Use the FORTRAN indexing (same as documentation) to
        # set and access info/cntl arrays from Python, whereas C
        # functions use C indexing. Maybe this is too confusing.
        self.lib.set_icntl(self._ma57, i - 1, val)

    def get_icntl(self, i):
        validate_index(i, self.icntl_len, 'ICNTL')
        return self.lib.get_icntl(self._ma57, i - 1)

    def set_cntl(self, i, val):
        validate_index(i, self.cntl_len, 'CNTL')
        validate_value(val, float, 'CNTL')
        self.lib.set_cntl(self._ma57, i - 1, val)

    def get_cntl(self, i):
        validate_index(i, self.cntl_len, 'CNTL')
        return self.lib.get_cntl(self._ma57, i - 1)

    def get_info(self, i):
        validate_index(i, self.info_len, 'INFO')
        return self.lib.get_info(self._ma57, i - 1)

    def get_rinfo(self, i):
        validate_index(i, self.rinfo_len, 'RINFO')
        return self.lib.get_info(self._ma57, i - 1)

    def do_symbolic_factorization(self, dim, irn, jcn):
        irn = irn.astype(np.intc, casting='safe', copy=True)
        jcn = jcn.astype(np.intc, casting='safe', copy=True)
        # TODO: maybe allow user the option to specify size of KEEP
        ne = irn.size
        self.ne_cached = ne
        self.dim_cached = dim
        assert ne == jcn.size, 'Dimension mismatch in row and column arrays'
        self.lib.do_symbolic_factorization(self._ma57, dim, ne, irn, jcn)
        return self.get_info(1)

    def do_numeric_factorization(self, dim, entries):
        entries = entries.astype(np.float64, casting='safe', copy=True)
        ne = entries.size
        assert ne == self.ne_cached, (
            'Wrong number of entries in matrix. Please re-run symbolic'
            'factorization with correct nonzero coordinates.'
        )
        assert dim == self.dim_cached, (
            'Dimension mismatch between symbolic and numeric factorization.'
            'Please re-run symbolic factorization with the correct '
            'dimension.'
        )
        if self.fact_factor is not None:
            min_size = self.get_info(9)
            self.lib.alloc_fact(self._ma57, int(self.fact_factor * min_size))
        if self.ifact_factor is not None:
            min_size = self.get_info(10)
            self.lib.alloc_ifact(self._ma57, int(self.ifact_factor * min_size))

        self.lib.do_numeric_factorization(self._ma57, dim, ne, entries)
        return self.get_info(1)

    def do_backsolve(self, rhs, copy=True):
        rhs = rhs.astype(np.double, casting='safe', copy=copy)
        shape = rhs.shape
        if len(shape) == 1:
            rhs_dim = rhs.size
            nrhs = 1
            rhs = np.array([rhs])
        elif len(shape) == 2:
            # FIXME
            raise NotImplementedError(
                'Functionality for solving a matrix of right hand '
                'is buggy and needs fixing.'
            )
            rhs_dim = rhs.shape[0]
            nrhs = rhs.shape[1]
        else:
            raise ValueError('Right hand side must be a one or two-dimensional array')
        # This does not necessarily need to be true; each RHS could have length
        # larger than N (for some reason). In the C interface, however, I assume
        # that LRHS == N
        assert self.dim_cached == rhs_dim, 'Dimension mismatch in RHS'
        # TODO: Option to specify a JOB other than 1. By my understanding,
        # different JOBs allow partial factorizations to be performed.
        # Currently not supported - unclear if it should be.

        if nrhs > 1:
            self.lib.set_nrhs(self._ma57, nrhs)

        if self.work_factor is not None:
            self.lib.alloc_work(self._ma57, int(self.work_factor * nrhs * rhs_dim))

        self.lib.do_backsolve(self._ma57, rhs_dim, rhs)

        if len(shape) == 1:
            # If the user input rhs as a 1D array, return the solution
            # as a 1D array.
            rhs = rhs[0, :]

        return rhs



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os.path
import subprocess

import pyomo.common
from pyomo.common.tempfiles import TempfileManager
from pyomo.common.errors import ApplicationError

from pyomo.opt.base import ProblemFormat, ConverterError


class PicoMIPConverter:
    def can_convert(self, from_type, to_type):
        """Returns true if this object supports the specified conversion"""
        #
        # Test if the glpsol executable is available
        #
        if not pyomo.common.Executable("pico_convert"):
            return False
        #
        # Return True for specific from/to pairs
        #
        if from_type == ProblemFormat.nl and to_type == ProblemFormat.cpxlp:
            return True
        if from_type == ProblemFormat.nl and to_type == ProblemFormat.mps:
            return True
        if from_type == ProblemFormat.mps and to_type == ProblemFormat.cpxlp:
            return True
        if from_type == ProblemFormat.cpxlp and to_type == ProblemFormat.mps:
            return True
        return False

    def available(self):
        return pyomo.common.Executable("pico_convert").available()

    def apply(self, *args, **kwargs):
        """
        Run the external pico_convert utility
        """
        if len(args) != 3:
            raise ConverterError(
                "Cannot apply pico_convert with more than one filename or model"
            )
        _exe = pyomo.common.Executable("pico_convert")
        if not _exe:
            raise ConverterError("The 'pico_convert' application cannot be found")

        pico_convert_cmd = _exe.path()
        target = str(args[1])
        if target == "cpxlp":
            target = "lp"
        # NOTE: if you have an extra "." in the suffix, the pico_convert program fails to output to the correct filename.
        output_filename = TempfileManager.create_tempfile(
            suffix='pico_convert.' + target
        )
        if not isinstance(args[2], str):
            fname = TempfileManager.create_tempfile(
                suffix='pico_convert.' + str(args[0])
            )
            args[2].write(filename=fname, format=args[1])
            cmd = (
                pico_convert_cmd
                + " --output="
                + output_filename
                + " "
                + target
                + " "
                + fname
            )
        else:
            cmd = pico_convert_cmd + " --output=" + output_filename + " " + target
            for item in args[2:]:
                if not os.path.exists(item):
                    raise ConverterError("File " + item + " does not exist!")
                cmd = cmd + " " + item
        print("Running command: " + cmd)
        subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if not os.path.exists(output_filename):  # pragma:nocover
            raise ApplicationError(
                "Problem launching 'pico_convert' to create " + output_filename
            )
        return (output_filename,), None  # no variable map at the moment
